let show_ctx_flag = Flags.show_ctx

let debug = Flags.tcc_debug


module A = Ast
module HS = HereditarySubstitution


type uvar_tp = 
  UTpVar | UTmVarRec of A.id list (* list of rec constants to be avoided  *)
         | UTmVarContrac of A.id list(* list of rec constants to be avoided *)

type uvar_res = 
Unresolved 
| ResolvedTp of A.tp (* contextual modal type modeled by pi abstractions *)
| ResolvedTm of (A.tm * A.tp) (* also provides type as an index, the type is index only for the moment*)

type assump = ((A.id * A.id list) * (A.id * A.id list)) list
type ctxname = 
Aid of A.id * A.tp
| FreeVar of string * A.tp
| UniVar of string * uvar_tp * uvar_res (* assumed to be unique *)
(* | WeakTpVar of A.id * A.weaktp used only during unification *)
| UnifContraintTp of A.tp * A.tp
| UnifContraintTm of A.tm * A.tm * A.tp  (* index *)
                                * assump (* assumption*)
| RecDef of A.id * A.tp * A.tm

type id_classification = IdBound | IdFreeVar | IdUniVar of uvar_tp | IdConstant of int (* number of implicit *)
                        | IdRecConstantInSig  of A.tm * A.rec_check_state
                        | IdRecConstantInCtx  of A.tm
type ctxtype = (ctxname) list

let new_metavar_tp (ctx : ctxtype) :  ctxtype * string = 
  let name = Uid.nextu() in
  let ctx = (UniVar (name, UTpVar, Unresolved)):: ctx in
  (ctx, name)
let new_metavar_tm (ctx : ctxtype) (tp : A.tp) (rec_avoids : A.id list) :  ctxtype * string = 
  let name = Uid.nextu() in
  let ctx = (Aid(A.Gen name, tp)) :: (UniVar (name, UTmVarRec rec_avoids, Unresolved)):: ctx in
  (ctx, name)
let new_metavar_tm_contrac (ctx : ctxtype) (tp : A.tp)  (rec_avoids : A.id list):  ctxtype * string = 
  let name = Uid.nextu() in
  let ctx = (Aid(A.Gen name, tp)) :: (UniVar (name, UTmVarContrac rec_avoids, Unresolved)):: ctx in
  (ctx, name)

  


let ds_uvar_res (res : uvar_res) : string = 
  match res with
  | Unresolved -> "unresolved"
  | ResolvedTm (tm, idx) -> "(resolved as " ^ A.ds_tm tm ^ " of type " ^ A.ds_tp idx ^ ")"
  | ResolvedTp (tp) -> "(resolved as " ^ A.ds_tp tp ^ ")"

let ds_uvar_tp (tp : uvar_tp) : string = 
  match tp with
  | UTmVarRec l -> "[tmr(" ^ A.ds_ids l ^ ")]"
  | UTmVarContrac l -> "[tmc(" ^ A.ds_ids l ^ ")]"
  | UTpVar -> "[tp]"

let ds_id_cls (cls :id_classification ) : string = 
  match cls with
  | IdBound -> "[cls:IdBount]"
  | IdFreeVar -> "[cls:IdFreeVar]"
  | IdUniVar utp -> "[cls:IdUniVar:" ^ ds_uvar_tp utp ^ "]"
  | IdConstant nimp -> "[cls:IdConstant:" ^ Int.to_string nimp ^ "]"
  | IdRecConstantInSig(tm,st) -> "[cls:IdRecConstantInSig:" ^ A.ds_tm tm ^ A.ds_rec_check_state st ^ "]"
  | IdRecConstantInCtx tm -> "[cls:IdRecConstantInCtx:" ^ A.ds_tm tm ^ "]"

let ds_assump_comp ((h,s) : A.id * A.id list) : string = 
  A.ds_tm (A.Atm(h, List.map(fun v -> A.Atm(v, [])) s))
let ds_assump (assump : assump) : string = 
  "[" ^ String.concat ", "  (List.map (fun (m1, m2) -> ds_assump_comp m1 ^ " = " ^ ds_assump_comp m2) assump) ^ "]"
let ds_ctxname (n : ctxname) : string = 
  match n with
  | Aid(n,p) -> "[id]" ^ A.ds_id n ^ " : " ^ A.ds_tp p
  | FreeVar(n,p) -> "[F]" ^ n ^ " : " ^ A.ds_tp p
  | UniVar(n,tp, r) -> "[U]" ^ ds_uvar_tp tp ^ n ^ " -> " ^ ds_uvar_res r
  | UnifContraintTm(m,n,t,assump) -> "[=tm]" ^ A.ds_tm m ^ " = " ^ A.ds_tm n ^ " : " ^ A.ds_tp t ^ ds_assump assump
  | UnifContraintTp(m,n) -> "[=tp]" ^ A.ds_tp m ^ " = " ^ A.ds_tp n
  | RecDef(n, tp, tm) -> "[=def]" ^ A.ds_id n ^ " : " ^ A.ds_tp tp ^ " = " ^ A.ds_tm tm

let prev_ctx : ctxtype ref = ref []
let ds_reset_ctx_printing () : unit = 
  prev_ctx := []

let ds_ctxtype (ctx : ctxtype) : string = 
  let addition = List.filter (fun cn -> not (List.mem cn (!prev_ctx))) ctx in
  let removal = List.filter (fun cn -> not (List.mem cn (ctx))) (!prev_ctx) in
  let addition_text = List.map (fun cn -> "+ " ^ ds_ctxname cn) addition  in
  let removal_text = List.map (fun cn -> "- " ^ ds_ctxname cn) removal  in
  let full_ctx = List.map ds_ctxname ctx in
  let prev = !prev_ctx in
  let _ = prev_ctx := ctx in
  if Flags.ctx_always_show_full then
  "{" ^ String.concat ",\n\t\t\t\t\t\t\t\t\t\t " full_ctx ^ "} " 
  ^ ("(full ctx)")
  else 
  "{" ^ String.concat ",\n\t\t\t\t\t\t\t\t\t\t " (removal_text @ addition_text) ^ "} " 
  ^ (if List.length prev <> 0 then "(changes only)" else "(full ctx)")


let ds_conditional_show_ctx (ctx : ctxtype) : string = 
  if show_ctx_flag 
    then " \tin context " ^ ds_ctxtype ctx
    else ""

let rec remove_first_name (ctx : ctxtype) (name : A.id) (isweak : bool) : ctxtype =
  match ctx with
  | [] -> failwith "attempt to remove first name from an empty context (or hereditarily empty)"
  | (((FreeVar _ | UniVar _ | RecDef _) as h):: t) -> h :: remove_first_name t name isweak
  | (((UnifContraintTm _) as h):: t) -> h :: remove_first_name t name isweak
  | (((UnifContraintTp _) as h):: t) -> h :: remove_first_name t name isweak
  (* | (((WeakTpVar (s, _))  as h):: t) -> 
    if s = name && isweak then t else 
              h :: remove_first_name t name isweak *)

  | (((Aid (s, _)) as h):: t) -> if s = name && not isweak then t else 
              h :: remove_first_name t name isweak
  

let with_new_name (ctx : ctxtype) (name : A.id) (tp : A.tp) 
  (next : ctxtype -> (ctxtype * 'a) ) : (ctxtype * 'a) = 
  match name with
  | A.Placeholder -> next ctx (* do not insert placeholder into context *)
  | A.Gen _ -> failwith "Cannot insert name with unification variables"
  | A.Prvd _ -> let (newctx, res) = next (Aid (name, tp) :: ctx) in
                (remove_first_name newctx name false, res) 

(* let with_new_name_weaktp (ctx : ctxtype) (name : A.id) (tp : A.weaktp) 
  (next : ctxtype -> (ctxtype * 'a) ) : (ctxtype * 'a) = 
  match name with
  | A.Placeholder -> failwith "cannot insert weak type with placeholder"
  | A.Gen _ -> failwith "Cannot insert name with unification variables"
  | A.Prvd _ -> let (newctx, res) = next (WeakTpVar (name, tp) :: ctx) in
                (remove_first_name newctx name true, res)  *)
let rec find_id_sig_global (ensure_tp : bool) (lookupid : string) (g: A.dec list) : (A.tp * id_classification ) option = 
  match g with
  | [] -> None
  | (A.Decl(_):: lt) -> find_id_sig_global ensure_tp lookupid lt
  | (A.Def(A.Prvd ca, tp, tm, st) :: lt) -> (* only find with checked signatures*)
        if ca = lookupid  
          then (
            if ensure_tp && st = A.Unchecked 
              then failwith "definitions can only be forward referenced in defintiion body, encoutered 
              a definition with same name but is unchecked. "
              else
                Some(tp, IdRecConstantInSig(tm, st)) 
              (* then  *)
          )
      else find_id_sig_global ensure_tp lookupid lt
  | (A.Def(_):: _) -> failwith ("malformed global signature (expects recursion constants to have a name): \n" ^ A.ds_decl g)

let rec find_id_sig_local (ensure_tp : bool) (lookupid : string) ((local, g): A.lgsig) : (A.tp * id_classification ) option = 
  match local with
  | (A.Decl(A.Prvd ca, tp, Some n_implicit):: lt) -> 
        if ca = lookupid then Some(tp, IdConstant n_implicit) else find_id_sig_local ensure_tp lookupid (lt, g)
  | (A.Decl(_):: _) -> failwith ("malformed local signature [decl]: \n" ^ A.ds_decl local)
  | (A.Def(A.Prvd ca, tp, tm, ((A.CheckedTp _ | A.CheckedBody _) as st)) :: lt) ->
        if ca = lookupid then Some(tp, IdRecConstantInSig(tm, st)) else find_id_sig_local ensure_tp lookupid (lt, g)
  | (A.Def(_):: _) -> failwith ("malformed local signature [def]: \n" ^ A.ds_decl local)
  | [] -> ( find_id_sig_global ensure_tp lookupid g)(* try looking up in global sig (remaining)*)

let rec lookup_metavar_type (ctx : ctxtype) (uvar_name : string)  : uvar_tp = 
  match ctx with
  | [] -> failwith ("metavar " ^ uvar_name ^ " not found")
  | (UniVar(n, tp, _):: ct) -> if n = uvar_name then tp else lookup_metavar_type ct uvar_name
  | (_ :: ct ) -> lookup_metavar_type ct uvar_name

let rec find_id_in_ctx (ctx : ctxtype) (name : A.id) : (A.tp * id_classification) option = 
  match name with
  | A.Placeholder -> failwith "cannot lookup placeholder "
  | (A.Prvd lookupid | A.Gen lookupid) ->
    (
      match ctx with
      | [] -> None
      | (Aid ( A.Prvd id, tp):: t) -> if id = lookupid then Some (tp, IdBound) else find_id_in_ctx t name 
      | (Aid ( A.Gen id, tp):: t) -> if id = lookupid then Some (tp, IdUniVar (lookup_metavar_type ctx id)) else find_id_in_ctx t name 
      | (Aid (_, _):: _) -> failwith "malformed context"
      | (RecDef(A.Prvd id, tp, tm) :: t) -> if id = lookupid then Some (tp, IdRecConstantInCtx(tm)) else find_id_in_ctx t name 
      | (RecDef(_) :: _) ->  failwith "tcc162: malformed context"
      | (FreeVar (id, tp) :: t ) -> if id = lookupid then Some (tp, IdFreeVar) else find_id_in_ctx t name 
      | ((UniVar _) :: t) -> find_id_in_ctx t name 
      | ((UnifContraintTm _) :: t) -> find_id_in_ctx t name 
      | ((UnifContraintTp _) :: t) -> find_id_in_ctx t name 
    )

(* Looks up free variables and bound variables, does not lookup unification variables nor weak variables *)
let find_id_in_ctx_and_sig (ctx : ctxtype) (ensure_tp : bool) (name : A.id)(lg : A.lgsig) : (A.tp * id_classification) option = 
  match name with
  | A.Placeholder -> failwith "cannot lookup placeholder "
  | (A.Prvd lookupid | A.Gen lookupid) ->
      (match find_id_in_ctx ctx name with
        | Some v -> Some v
        | None -> find_id_sig_local ensure_tp lookupid lg 
      )

(* ensure_tp triggers an error if looked up is a constant whose id has not been checked*)
let  lookup_id_in_ctx (ctx : ctxtype)  (name : A.id) : (A.tp * id_classification) = 
  match find_id_in_ctx ctx name with
  | Some t -> t
  | None -> failwith ("tcc85: lookup not found" ^ "\n when looking up " ^ A.ds_id name ^ " in context " ^ ds_conditional_show_ctx ctx)

let  lookup_id_in_ctx_and_sig (ctx : ctxtype) (ensure_tp : bool) (name : A.id)(lg : A.lgsig) : (A.tp * id_classification) = 
  match find_id_in_ctx_and_sig ctx ensure_tp name lg  with
  | Some t -> t
  | None -> failwith ("tcc85: lookup not found" ^ "\n when looking up " ^ A.ds_id name ^ " in context " ^ ds_conditional_show_ctx ctx)

(* presupposes id is found all unfound ids will be classified as Const*)
(* let rec classify_id (ctx : ctxtype) (name : A.id)(lg : A.lgsig) : id_classification = 
  match name with
  | A.Placeholder -> failwith "cannot classify placeholder "
  | A.Gen _ -> IdUniVar
  | A.Prvd lookupid ->
    (
      match ctx with
      | [] -> IdConstant
      | (Aid ((A.Prvd id | A.Gen id), _):: t) -> if id = lookupid then IdBound else classify_id t name lg
      | (Aid (_, _):: _) -> failwith "malformed context"
      | (FreeVar (id, _) :: t ) -> if id = lookupid then IdFreeVar else classify_id t name lg
      | ((UniVar _) :: t) -> classify_id t name lg
      | ((WeakTpVar ((A.Prvd id | A.Gen id), _)) :: t) -> 
              if id = lookupid then IdBound else classify_id t name lg
      | ((WeakTpVar (_)) :: _) -> failwith "malformed context?"
    ) *)

(* this method looks up weaktp, antique because now weaktypes are removed from contexts*)
let lookup_id_weaktp (ctx : ctxtype) (name : A.id) (lg : A.lgsig) : (A.weaktp * id_classification) = 
  let (tp, cls) = lookup_id_in_ctx_and_sig ctx true name lg in (A.to_weak tp, cls)

  (* match name with
  | A.Placeholder -> failwith "cannot lookup placeholder "
  (* | A.Gen _ -> failwith "cannot lookup generated name " *)
  | (A.Prvd lookupid | A.Gen lookupid)->
    (
      match ctx with
      | [] -> 
        (match find_id_sig_local lookupid lg with
        | None -> failwith ("tcc88: lookup not found when looking up " ^ A.ds_id name)
        | Some(tp, cls) -> (A.to_weak tp, cls))
      | (Aid ((A.Prvd id | A.Gen id), tp):: t) -> if id = lookupid then (A.to_weak tp, IdBound) else lookup_id_weaktp t name lg
      | (Aid (_, _):: _) -> failwith "malformed context"
      | (FreeVar (id, tp) :: t ) -> if id = lookupid then  (A.to_weak tp, IdFreeVar) else lookup_id_weaktp t name lg
      | ((UniVar _ | UnifContraintTm _ | UnifContraintTp _   ) :: t) -> lookup_id_weaktp t name lg
      (* | ((WeakTpVar ((A.Prvd id | A.Gen id), wtp)) :: t) -> 
              if id = lookupid then (wtp, IdBound) else lookup_id_weaktp t name lg
      | ((WeakTpVar (_)) :: _) -> failwith "malformed context?" *)
    ) *)
    
    
  

(* checks whether the string begins with an upper case letter *)
let str_is_upper (s : string) = 
  let firstLetter = String.sub s 0 1 in
  String.uppercase_ascii firstLetter = firstLetter

(* collects all bound variables, do not collect unification variables with type declaration*)
(* this behavior is consistent with the current implementation of Twelf *)
let rec collect_all_bound_vars (ctx : ctxtype) : (A.id * A.tp) list = 
  match ctx with
  | [] -> []
  | (Aid(A.Prvd n, t) :: ct) ->  collect_all_bound_vars ct @ [(A.Prvd n , t)]
  | _ :: ct -> collect_all_bound_vars ct

(* collects all bound and free vars in that order *)
let rec collect_all_bound_and_free_vars (ctx : ctxtype) : (A.id * A.tp) list = 
  match ctx with
  | [] -> []
  | (Aid(A.Prvd n, t) :: ct) ->  collect_all_bound_and_free_vars ct @ [(A.Prvd n , t)]
  | (FreeVar(n, t) :: ct) ->  collect_all_bound_and_free_vars ct @ [(A.Prvd n , t)]
  | _ :: ct -> collect_all_bound_and_free_vars ct

let rec collect_all_free_vars (ctx : ctxtype) : (A.id * A.tp) list = 
  match ctx with
  | [] -> []
  | (FreeVar(n, t) :: ct) ->  collect_all_free_vars ct @ [(A.Prvd n , t)]
  | _ :: ct -> collect_all_free_vars ct

let rec collect_all_rec_constants (ctx : ctxtype) : (A.id) list = 
  match ctx with
  | [] -> []
  | (RecDef(n, _,_) :: ct) ->  collect_all_rec_constants ct @ [n]
  | _ :: ct -> collect_all_rec_constants ct



(* abstract_over G D T = \D. T *)
let abstract_over_tp (ctx : ctxtype) (ids : A.id list) (base : A.tp) (lg : A.lgsig) : A.tp = 
  List.fold_right (fun id acc -> 
      let lookup = find_id_in_ctx_and_sig ctx true id lg in
      match lookup with
      | Some (tp, (IdBound | IdFreeVar)) -> A.Pi(id, tp, acc)
      | Some _ -> failwith "tcc96: cannot abstract over non-bound variables"
      | None  -> failwith "tcc97: name not found in context"
    ) ids base
let abstract_over_tm (_ : ctxtype) (ids : A.id list) (base : A.tm) (_ : A.lgsig) : (A.tm ) = 
  List.fold_right (fun id (acctm) -> 
      (A.Lam(id, acctm))
    ) ids base

let abstract_over_tm_tp (ctx : ctxtype) (ids : A.id list) (base : A.tm * A.tp) (lg : A.lgsig) : (A.tm * A.tp) = 
  List.fold_right (fun id (acctm, acctp) -> 
      let (lookuptp, _) = lookup_id_in_ctx_and_sig ctx true id lg in
      (A.Lam(id, acctm), A.Pi(id, lookuptp, acctp))
    ) ids base

let abstract_over_tm_weaktp (ctx : ctxtype) (ids : A.id list) (base : A.tm * A.weaktp) (lg : A.lgsig): (A.tm * A.weaktp) = 
  List.fold_right (fun id (acctm, acctp) -> 
      let (lookuptp, _) = lookup_id_weaktp ctx id lg in
      (A.Lam(id, acctm), A.Arr(lookuptp, acctp))
    ) ids base

(* creates a term awaiting unification
    u * (x1, x2, ..., xn), with a fresh u (unification varaible of appropriate types),
    xi's are provided by an extraction function, i.e. one of 
    collect_all_(bound/free/bound_and_free)_vars
    *)
let new_metavar_tm_with_abstraction (ctx : ctxtype) 
(extraction: ctxtype -> (A.id * A.tp) list) 
(tp_base : A.tp) (rec_avoids : A.id list)
(lg : A.lgsig) :  ctxtype * A.tm = 
  let vars = extraction ctx in
  let ids = List.map (fun (id, _) -> id) vars in
  let uvar_tp = abstract_over_tp ctx ids tp_base lg in
  let (ctx, uname) = new_metavar_tm ctx uvar_tp rec_avoids in
  let tm = A.Atm(A.Gen(uname), (List.map (fun (id, tp) -> HS.eta_expand (A.Atm(id, [])) (A.to_weak tp)) vars)) in
  (ctx, tm)

(* instantiates Pi x: A.B with (u y_1 ... y_n), 
   returns (ctx, u y_1 ... y_n, [u y_1 ... y_n/x]B)*)
let instantiate_pi_with_new_metavar (ctx : ctxtype) 
(extraction: ctxtype -> (A.id * A.tp) list) 
(tp : A.tp) (rec_avoids : A.id list)
(lg : A.lgsig) :  ctxtype * A.tm * A.tp = 
match tp with
| A.Pi(x, t2, t1) -> 
    let (ctx, tm) = new_metavar_tm_with_abstraction ctx extraction t2 rec_avoids lg in
    let t1 = HS.h_sub_in_tp (HS.eta_expand (tm) (A.to_weak t2))
                (A.to_weak t2) x t1 in
    (ctx, tm, t1)
|  _ -> failwith "expecting pi types"

(* generates metavars for use in Projection/Imitation 
   the generated type is assumed to be applied to abstraction ids, 
  will generate length length_spine many uvars*)
let new_metavars_tm (ctx : ctxtype) (h2tp : A.tp) (length_spine : 'a list) (abstraction_ids : A.id list) 
  (rec_avoids : A.id list) (lg : A.lgsig) :  ctxtype * string list = 
  let (ctx, unifVars, _) = List.fold_left (fun (ctx, unifVars, tp) _ -> 
    match tp with
    | A.Type _ | A.CoType _ -> failwith "unif187: Should not be Type/CoType"
    | A.Atp _ -> failwith "unif188: not enough pi types to destruct"
    | A.Pi(x, t2, t1) -> 
      let t2' = abstract_over_tp ctx abstraction_ids t2 lg in
      let (ctx, newVar) = new_metavar_tm ctx t2' rec_avoids in
      let expandedTm = HS.eta_expand (A.Atm(A.Gen newVar, [])) (A.to_weak t2) in
      let t1 =  HS.h_sub_in_tp expandedTm (A.to_weak t2) x t1 in
      (ctx, unifVars@[newVar], t1)
  ) (ctx, [], h2tp) length_spine in
    (ctx, unifVars)


(* name may be modified during resolution*) (* int denotes number of implicit arguments (in case looking up in signature) *)
(* returns the resolved id h, along with a list of arguments s, the type returned will be the type of Atm(h, s)*)
let resolve_id (ctx : ctxtype)(name : A.id) (lg : A.lgsig) : ctxtype * A.id * (A.tp * id_classification) = 
    match name with
    | A.Placeholder -> (* insert new unification variable with unification type *)  
    failwith "tcc243: resolve_id does not support synthesis of placeholers, please check directly"
       
    | A.Gen s -> 
      (match find_id_in_ctx_and_sig ctx true name lg with
      | Some tp -> (ctx, name, tp)
      | None -> failwith ("tcc248: name " ^ s ^ " not found in context"))
      (* failwith "cannot resolve generated vars" *)
    | A.Prvd s -> (
      match find_id_in_ctx_and_sig ctx true name lg with
      | Some tp -> (ctx, name, tp)
      | None -> (
        if str_is_upper s 
        then failwith "metavars should have been bound at the outside"
                    (* (let (ctx, tpname) = new_metavar_tp ctx in (* TODO: Figure this out, 
          SOLUTION is perhaps: apparently since freevars are bound at the outside, 
            its type should not involve anything in its scope*)
          let vars = collect_all_free_vars ctx in
          let tml = List.map (fun (v, t) -> HS.eta_expand (A.Atm(v, [])) (A.to_weak t) ) vars in
          let tp = A.Atp(A.Gen tpname, tml) in
          let ctx =  FreeVar (s, tp) :: ctx in
          (ctx, name, (tp, IdFreeVar))) *)
        else (failwith ("tcc68: name " ^ s ^ " not found in context"))
      )
    )

let rec lookup_metavar (ctx : ctxtype) (uvar_name : string)  : uvar_res = 
  match ctx with
  | [] -> failwith ("metavar " ^ uvar_name ^ " not found")
  | (UniVar(n, _, res):: ct) -> if n = uvar_name then res else lookup_metavar ct uvar_name
  | (_ :: ct ) -> lookup_metavar ct uvar_name


let subst_uvar_in_tm (n : uvar_res) (x : A.id) (m : A.tm) : A.tm = 
  match n with
  | Unresolved -> failwith "tcc102: cannot subst unresolved uvar"
  | ResolvedTm (n, idx) -> HS.h_sub_in_tm n (A.to_weak idx) x m
  | ResolvedTp _ -> m
let subst_uvar_in_tp (n : uvar_res) (x : A.id) (m : A.tp) : A.tp = 
  match n with
  | Unresolved -> failwith "tcc102: cannot subst unresolved uvar"
  | ResolvedTm (n, idx) -> HS.h_sub_in_tp n (A.to_weak idx) x m
  | ResolvedTp p -> HS.h_sub_tp_in_tp p x m

let subst_uvar_in_weaktp (n : uvar_res) (x : A.id) (t : A.weaktp) : A.weaktp = 
  match n with
  | Unresolved -> failwith "tcc102: cannot subst unresolved uvar"
  | ResolvedTm _ -> t
  | ResolvedTp p -> HS.h_sub_tp_in_weaktp p x t


let subst_uvar_in_uvar (n : uvar_res) (x : A.id) (m : uvar_res) = 
  (match m with
  | Unresolved  -> Unresolved
  | ResolvedTm (m, p) -> ResolvedTm(subst_uvar_in_tm n x m, subst_uvar_in_tp n x p)
  | ResolvedTp p -> ResolvedTp(subst_uvar_in_tp n x p)
  )
  
let resolve_metavar (ctx : ctxtype) (uvar_name : string) (res : uvar_res) : ctxtype = 
  let _ = if debug then print_endline ("resolve_metavar resolving " ^ uvar_name ^ " as " ^ ds_uvar_res res ^ " in ctx " ^ ds_conditional_show_ctx ctx) else () in
  let present = ref false in
  let rec recur ctx = 
  (match ctx with
  | [] -> [] 
  | (UniVar(n, utp, r):: ct) -> if n = uvar_name 
                            then  (match r with
                                    | Unresolved -> (let _ = present := true in
                                    UniVar(n,utp, res) :: recur ct )
                                    | _ -> 
                                      failwith ("tcc100: Already resolved metavar " ^ n ^ " as " ^ ds_uvar_res r 
                                    ^ "\n when attempting to resolve " ^ uvar_name ^ " as " ^ ds_uvar_res res))
                            else 
                              ( UniVar(n,utp,
                              subst_uvar_in_uvar res (A.Gen uvar_name) r
                              )) :: recur ct 
  | ((FreeVar (name, tp)) :: ct ) -> 
    FreeVar (name, subst_uvar_in_tp res (A.Gen uvar_name) tp) ::
    recur ct
  | ((Aid (name, tp)) :: ct ) -> 
    Aid (name, subst_uvar_in_tp res (A.Gen uvar_name) tp) ::
    recur ct 
  | ((RecDef (name, tp, tm)) :: ct ) -> 
    RecDef (name, subst_uvar_in_tp res (A.Gen uvar_name) tp, subst_uvar_in_tm res (A.Gen uvar_name) tm) ::
    recur ct 
  (* | ((WeakTpVar(name, t) :: ct)) ->  (WeakTpVar (name, subst_uvar_in_weaktp res (A.Gen uvar_name) t)) :: recur ct *)
  | ((UnifContraintTm(m1, m2, t, assump)) :: ct)  -> 
          (UnifContraintTm(subst_uvar_in_tm res (A.Gen uvar_name) m1, subst_uvar_in_tm res (A.Gen uvar_name) m2, 
          subst_uvar_in_tp res (A.Gen uvar_name) t, assump)) :: recur ct
  | ((UnifContraintTp(t1, t2)) :: ct)  -> 
          (UnifContraintTp(subst_uvar_in_tp res (A.Gen uvar_name) t1, subst_uvar_in_tp res (A.Gen uvar_name) t2)) :: recur ct
  ) in
  let ctx = recur ctx in
  let _ = if !present = false then failwith "tcc158: resolution of univar not present in context" else ()
in
  (* let _ = if debug then  print_endline ("resolved metavar " ^ uvar_name ^ " as " ^ ds_uvar_res res ^ " producing ctx " ^ ds_conditional_show_ctx ctx) else () in *)
  ctx

  

(* this is a hack, because metavars do not actually behave like references, 
generated types may contain metavariables that have been resolved but not substituted.
This procedure go through the types and instantiate all resolved metavaraibles
*)
let instantiate_metavar_tp (ctx : ctxtype) ( t : A.tp)  : A.tp = 
  List.fold_right (fun ctxitem t -> 
      match ctxitem with
      | UniVar(_,_,  Unresolved) -> t
      | UniVar(s,_, r) -> subst_uvar_in_tp r (A.Gen s) t
      | _ -> t
      )  ctx t

(* replace all metavars in tm with their instantiations in ctx *)
let instantiate_metavar_tm (ctx : ctxtype) ( m : A.tm)  : A.tm = 
  List.fold_right (fun ctxitem m -> 
      match ctxitem with
      | UniVar(_,_,  Unresolved) -> m
      | UniVar(s,_, r) -> subst_uvar_in_tm r (A.Gen s) m
      | _ -> m
      )  ctx m
 
type inclusion_func_type = (string  (* the name of the identifier*)
* id_classification option  (* the classification of the identifier, None if identifier not found in context and signature*)
* bool (*true if is in tm *) 
* bool (* true if it is locally bound (has type __get_heads_local_tp__marker___ )*) 
  ->  bool)

let get_heads_local_bound_tp = A.Atp(A.Prvd "__get_heads_local_tp__marker___", [])
(* currently ignores unification heads, but may change TODO: decide *)
let rec get_heads_in_tm (ctx : ctxtype) 
(inclusion_func : inclusion_func_type)
 (tm : A.tm) (lg : A.lgsig) : ctxtype * string list = 
 try(
  match tm with
  | A.Atm(h, s) -> 
    let heads = 
        (
          match h with
          | A.Gen _ -> []
            (* failwith "cannot get freevars with gen" *)
          | A.Placeholder -> []
          | A.Prvd s -> 
          (match find_id_in_ctx_and_sig ctx false h lg  with
        | None -> 
          if inclusion_func (s, None, true, false) then [s] else [] 
        | Some (tp, cls) -> 
          if inclusion_func (s, Some cls, true, tp = get_heads_local_bound_tp) then [s] else [] 
          )
        ) in
    let (ctx, spines) = get_heads_in_spine ctx inclusion_func s lg in
    (ctx, heads @ spines)
  | A.Lam(x, b) ->  (* the type is irrelevant for free variables, so fill in empty types*)
  with_new_name ctx x get_heads_local_bound_tp (fun ctx -> 
      get_heads_in_tm ctx inclusion_func b lg
    )
 ) with Failure s -> (raise (Failure (s ^ "\n when getting heads in tm " ^ A.ds_tm tm)))

and get_heads_in_spine (ctx : ctxtype)
(inclusion_func : inclusion_func_type)
 (s : A.tm list) (lg : A.lgsig) : ctxtype * string list = 
  List.fold_left (fun (ctx, l) tm -> 
    let (ctx, fm) = get_heads_in_tm ctx inclusion_func tm lg in (ctx, l @ fm)) (ctx, []) s
let rec get_heads_in_tp (ctx : ctxtype)
(inclusion_func : inclusion_func_type)
 (tp : A.tp) (lg : A.lgsig) : ctxtype * string list = 
 try(
  match tp with
  | A.Type _ | A.CoType _ -> (ctx, [])
  | A.Pi(x, t2, t1) -> 
      let (ctx, t2f) = get_heads_in_tp ctx inclusion_func t2 lg in
      with_new_name ctx x t2 (fun ctx -> 
          let (ctx, t1f) = get_heads_in_tp ctx inclusion_func t1 lg 
          in (ctx, t2f @t1f)
        )
  | A.Atp( h, s) -> 
    let heads = 
      (
          match h with
          | A.Gen _ -> []
            (* failwith "cannot get freevars with gen" *)
          | A.Placeholder -> []
          | A.Prvd s -> 
      (match find_id_in_ctx_and_sig ctx false h lg  with
        | None -> 
          if inclusion_func (s, None, true, false) then [s] else [] 
        | Some (tp, cls) -> 
          if inclusion_func (s, Some cls, true, tp = get_heads_local_bound_tp) then [s] else [] 
      )
      )
    in
  let (ctx, spines) = get_heads_in_spine ctx inclusion_func s lg in
  (ctx, heads @ spines)
 ) with Failure s -> (raise (Failure (s ^ "\n when getting heads in tp " ^ A.ds_tp tp)))

let get_capital_free_vars_in_tp 
(tp : A.tp) (lg : A.lgsig) : A.id list 
= 
try(
  let (_, vars) = get_heads_in_tp [] (fun (s, cls, in_tm, _) -> 
      match cls with
      | None -> 
        (if str_is_upper s 
          then (if in_tm
            then true
            else failwith "tcc417: cannot have free variables range over types, types cannot be quantified over e.g. {j : type} is not allowed in LF"
            )
          else failwith ("tcc478: id not found. id is " ^ s)
        )
      | Some _ -> false
    ) tp lg in
  let vars = ListUtil.remove_duplicates vars in
      List.map (fun v -> A.Prvd v) vars
 ) with Failure s -> (raise (Failure (s ^ "\n when getting capital free vars in tp " ^ A.ds_tp tp)))


let get_bound_free_vars_in_tm (ctx : ctxtype)
  (tm : A.tm) (lg : A.lgsig) : A.id list 
= 
  try(
    let (_, vars) = get_heads_in_tm ctx (fun (s, cls, _, is_locally_bound) -> 
      match cls with
      | None -> 
          failwith ("tcc478: id not found. id is " ^ s)
      | Some (IdBound) -> if is_locally_bound then false else true
      | Some _ -> false
    ) tm lg in
    let vars = ListUtil.remove_duplicates vars in
        List.map (fun v -> A.Prvd v) vars
 ) with Failure s -> (raise (Failure (s ^ "\n when getting bound free vars in tm " ^ A.ds_tm tm)))
  
let get_rec_constants_in_tp 
(tp : A.tp) (lg : A.lgsig) : string list =
  try(
    let (_, consts) = 
    get_heads_in_tp [] (fun (_, cls, in_tm, _) -> 
      if not in_tm then false else
        match cls with
        | Some (IdRecConstantInSig _) -> true
        | Some (IdRecConstantInCtx _) -> failwith "unexpected rec constant in ctx"
        | _ -> false
      ) tp lg
      in consts
 ) with Failure s -> (raise (Failure (s ^ "\n when getting rec consts in tp " ^ A.ds_tp tp)))

let get_rec_constants_in_tm_ctx (ctx : ctxtype)
(tm : A.tm)  : A.id list =
  try(
  let (_, consts) = 
    get_heads_in_tm ctx (fun (_, cls, in_tm, _) -> 
    if not in_tm 
      then failwith "tcc544: not possible"
      else
      match cls with
      | Some (IdRecConstantInSig _) -> failwith "unexpected rec constant in sig"
      | Some (IdRecConstantInCtx _) -> true
      | _ -> false
    ) tm ([], []) in
  let consts = ListUtil.remove_duplicates consts in
      List.map (fun v -> A.Prvd v) consts
 ) with Failure s -> (raise (Failure (s ^ "\n when getting rec constant (using context only) in tm " ^ A.ds_tm tm)))

let get_rec_constants_in_tm_sig
(tm : A.tm) (lg : A.lgsig) : string list =
  try(
  let (_, consts) = 
    get_heads_in_tm [] (fun (_, cls, in_tm, _) -> 
    if not in_tm 
      then failwith "tcc544: not possible"
      else
      match cls with
      | Some (IdRecConstantInSig _) -> true
      | Some (IdRecConstantInCtx _) -> failwith "unexpected rec constant in ctx"
      | _ -> false
    ) tm lg
      in consts
 ) with Failure s -> (raise (Failure (s ^ "\n when getting rec constant (using context and sig) in tm " ^ A.ds_tm tm)))

(* prepopulate contexts with all variables that are free in tp*)
let populate_ctx_with_free_vars_from_tp (ctx : ctxtype) (tp : A.tp) (lg : A.lgsig) : ctxtype = 
  let (freeVars) = get_capital_free_vars_in_tp tp lg  in
  let freeVars = ListUtil.remove_duplicates freeVars in 
  let (ctx,tpVars) = List.fold_left (fun (ctx, tpVars) _ -> 
          let (ctx, tpname) = new_metavar_tp ctx in 
          (ctx, tpVars@[tpname])
    ) (ctx, []) freeVars in
  let tpl = List.map (fun (h) -> (A.Atp(A.Gen h, 
        List.map (fun arg -> A.Atm( arg, [])) (List.filteri (fun i _ -> List.nth tpVars i <> h) freeVars)))) tpVars  in
  let (ctx) = List.fold_left2 (fun ctx free tp  ->
    match free with
    | A.Prvd free -> FreeVar(free, tp) :: ctx
    | _ -> failwith "expect frvd"
  ) ctx freeVars tpl
  in ctx

let rec populate_ctx_with_rec_constants_from_signature (ctx : ctxtype) (l : A.dec list) : ctxtype = 
  match l with
  | [] -> ctx
  | (A.Def(_, _, _, A.CheckedBody (l)) :: t) -> 
      populate_ctx_with_rec_constants_from_signature (
        (List.filter_map (fun (name, tp, tm) ->
          (* only add additional *)
          match find_id_in_ctx ctx name with
          | None -> Some(RecDef(name, tp, tm))
          | Some _ -> None
           ) l 
      )
           @ ctx) t
  | (_ :: t) -> populate_ctx_with_rec_constants_from_signature ctx t

let add_unification_equations_tp (ctx : ctxtype) (l : (A.tp * A.tp) list) : ctxtype = 
  List.map (fun (t1, t2) -> UnifContraintTp(t1, t2)) l @ ctx

let add_unification_equations_tm (ctx : ctxtype) (l : (A.tm * A.tm * A.tp * assump) list) : ctxtype = 
  List.map (fun (m1, m2, t, assump) -> UnifContraintTm(m1, m2, t, assump)) l @ ctx


(* Returns 1-1 *)
let rename_rec_consts_to_fresh_names (ctx : ctxtype) (to_rename : A.id list)  : (ctxtype * A.id list) = 
  let fresh_names = List.map (fun _ -> A.Prvd (Uid.nextr())) to_rename in
  let simultaneous renamef expr = List.fold_right2 (fun fresh old expr -> 
        renamef fresh old expr
      ) fresh_names to_rename expr in
  let rtp = simultaneous HS.rename_in_tp in (* renames function for tp *)
  let rtm = simultaneous HS.rename_in_tm in (* renames function for tm *)
  let rid = simultaneous (fun y x id -> if x = id then y else id) in (* renames function for id *)
  let ctx = List.map (fun ctxname -> 
    match ctxname with
      | Aid(x, tp) -> Aid(x, rtp tp)
      | FreeVar(x, tp) -> FreeVar(x, rtp tp)
      | UniVar(x, utp, ures) -> UniVar(x, utp, (match ures with
                      | ResolvedTp tp -> ResolvedTp (rtp tp)
                      | ResolvedTm (tm, tp) -> ResolvedTm (rtm tm, rtp tp)
                      | Unresolved -> Unresolved
                      ))
      | UnifContraintTm _ -> failwith "cannot rename during unificaiton"
      | UnifContraintTp _ -> failwith "cannot rename during unificaiton"
      | RecDef(x, tp, tm) -> RecDef(rid x, rtp tp, rtm tm)
    ) ctx in 
  (ctx, fresh_names)

(* this function 
upon receiving univar_name univar_args 
creates a new r with r = \univar_args. G^c univar_args, 
resolves univar_name with r, and return name G^c 
   *)
let turn_rec_univar_into_contra_univar (ctx : ctxtype) 
  (univar_name : string) (univar_args : A.id list) (lg : A.lgsig) : (ctxtype * A.id) = 
  let newR = A.Prvd( Uid.nextr()) in
  let (tp, cls) = lookup_id_in_ctx ctx (A.Gen univar_name) in
  let rec_avoids = (match cls with
    | IdUniVar (UTmVarRec l) -> l
    | _ -> failwith "tcc709 turning non recursive variable into contractive") in
  (* let (ctx, newUnivar) = new_metavar_tm_contrac ctx (abstract_over_tp ctx univar_args index lg)  in *)
  (* new recvar is always guarded (coinductive constructor will void the guard in the next step)*)
  let (ctx, newUnivar) = new_metavar_tm_contrac ctx tp (newR :: rec_avoids)  in
  let newRBody = A.Atm(A.Gen newUnivar, List.map (fun v -> A.Atm(v, [])) univar_args) in
  let (rdef) = abstract_over_tm ctx univar_args (newRBody) lg in
  let ctx = RecDef(newR, tp, rdef) :: ctx in
  let resolution_tm = abstract_over_tm ctx univar_args (A.Atm(newR, List.map (fun v -> A.Atm(v, [])) univar_args)) lg in
  let ctx = resolve_metavar ctx univar_name (ResolvedTm(resolution_tm, tp)) in
  (ctx, A.Gen newUnivar)

let to_rec_sig (ctx : ctxtype) (rec_consts : A.id list) : (A.rec_sig ) = 
  let _ = if ListUtil.contains_duplicate rec_consts then failwith "tcc691: requested sig contains duplicates" else ()  in
  let res = List.map (fun name -> 
      match find_id_in_ctx ctx name with
      | Some(tp, IdRecConstantInCtx def) -> (name, tp, def)
      | _ -> failwith "tcc695: not found"
    ) rec_consts
  in res

let get_avoid_list (ctx : ctxtype) (unifvar : A.id) = 
                    (
                      (match lookup_id_in_ctx ctx unifvar with
                      (_, IdUniVar (UTmVarRec l | UTmVarContrac l)) -> l
                      | _ -> failwith "unif153:")
                    )
  
(* check whether a type/kind is inductive or coinductive *)
(* let rec classify_tp_is_coind (tp : A.tp) (lg : A.lgsig) : bool = 
  match tp with
  | A.Atp(A.Prvd h, _) -> 
    (match find_id_sig_local true h lg with
    | None -> false
    | Some (tp, _) -> classify_tp_is_coind tp lg
    )
  | A.Atp(_, _) -> false (* generated name cannot *)
  | A.Pi(_, _, t1) -> classify_tp_is_coind t1 g
  | A.Type -> true
  | A.CoType -> false
                   *)