

module A = Ast
(* warning: this version of type checking considers all DefI to be coinductive, 
   and do not consider notational definitions *)

module Ctx = TypeCheckingContext
module HS = HereditarySubstitution
module Unif = Unification


let debug = Flags.tr_debug


let rec tr_kind (tr : A.tp) (ctx : Ctx.ctxtype) (lg : A.lgsig) : Ctx.ctxtype * A.tp = 
  let _ = if debug then print_endline ("tr_kind on " ^ A.ds_tp tr ^ Ctx.ds_conditional_show_ctx ctx ) else () in 
  try (match tr with
    | A.Type _ -> (ctx, tr)
    | A.CoType _ -> (ctx, tr)
    | A.Atp _ -> failwith "Should not check kind, is not kind"
    | A.Pi(name, t2, t1) -> 
        let (ctx, ct2) = tr_type_cotype t2 ctx lg in
        Ctx.with_new_name ctx name ct2 (fun ctx -> 
                let (ctx, ct1) = tr_kind t1 ctx lg in
            (ctx, A.Pi(name, ct2, ct1))
          )
  ) 
        with Failure s -> raise (Failure (s ^ "\n when checking " ^ A.ds_tp tr ^ " is a kind" ^ Ctx.ds_conditional_show_ctx ctx))
  


and tr_type_cotype (tr : A.tp) (ctx : Ctx.ctxtype) (lg : A.lgsig) : Ctx.ctxtype * A.tp = 
  let _ = if debug then print_endline ("tr_type_cotype on " ^ A.ds_tp tr ^ Ctx.ds_conditional_show_ctx ctx ) else () in 
    try(
      match tr with
      | A.Pi(x, t2, t1) -> 
          let (ctx, ct2) = tr_type_cotype t2 ctx lg in
          Ctx.with_new_name ctx x ct2 (fun ctx -> 
              let (ctx, ct1) = tr_type_cotype t1 ctx lg in 
              (ctx, A.Pi(x, ct2, ct1))
            )
      | A.Atp(A.Placeholder, []) -> 
          let (ctx, tpname) = Ctx.new_metavar_tp ctx in
          let vars = Ctx.collect_all_bound_and_free_vars ctx in
          let tml = List.map (fun (v, t) -> HS.eta_expand (A.Atm(v, [])) (A.to_weak t) ) vars in
          let tp = A.Atp(A.Gen tpname, tml) in
          (ctx, tp)
      | A.Atp(A.Placeholder, _) -> failwith "placeholder cannot be applied to arguments in the domain of types"
      | A.Atp _ ->  (* synth kind *)
          (let (ctx, ct, k) = tr_synth_kind tr ctx lg in
          match k with
            | A.Type _ -> (ctx, ct)
            | A.CoType _ -> (ctx, ct)
            | _ -> failwith ("not well typed, synthesized kind is not type/cotype, but is" ^ A.ds_tp k))
        | (A.Type _ | A.CoType _) -> failwith "tr_type_cotype fail: Type / CoType <= (co)type does not hold, Type/CoType is a kind, cannot quantify over kinds"
    ) with Failure s -> raise (Failure (s ^ "\n when checking that " ^ A.ds_tp tr ^ " is a (co)type"  ^ Ctx.ds_conditional_show_ctx ctx ))


and tr_synth_kind (tr : A.tp) (ctx : Ctx.ctxtype)   (lg : A.lgsig)
  : Ctx.ctxtype * A.tp (* resolved type *) * A.tp (* synthsized kind *) = 
  let _ = if debug then print_endline ("tr_synth_kind on " ^ A.ds_tp tr ^ Ctx.ds_conditional_show_ctx ctx ) else () in 
    try(
    match tr with
    | A.Pi _  -> failwith "cannot synthesize on pi types"
    | (A.Type _ | A.CoType _) -> failwith "cannot synthesize on kinds"
    | Atp(hd, spine) ->  (* synth kind *)
        let (ctx, resolvedname, resolvedtp) = Ctx.resolve_id ctx hd lg in 
          let (ctx, cspine, k) =  tr_feed_spine spine resolvedtp ctx lg in 
            (ctx, A.Atp(resolvedname, cspine), k)
    ) with Failure s -> raise (Failure (s ^ "\n when synthesizing the kind of " ^ A.ds_tp tr ^ ""  ^ Ctx.ds_conditional_show_ctx ctx ))
            
          

  
(* corresponds to the judgment S \rhd A/K => A/K *)
(* assumes fed is well-typed *)
and tr_feed_spine (spine : A.tm list) ((fed, tpcls) : A.tp * Ctx.id_classification) (ctx : Ctx.ctxtype)  (lg : A.lgsig) : Ctx.ctxtype * A.tm list *  A.tp = 
  let _ = if debug then print_endline ("tr_feed_spine on " ^ A.ds_spine spine ^ " into " ^ A.ds_tp fed ^ " in ctx " ^ Ctx.ds_conditional_show_ctx ctx ) else () in 
  (* this is another hack to ensure that all resolved metavars are propogated TODO: remove when metavars are truely references *)
  let fed = Ctx.instantiate_metavar_tp ctx fed in
  let n_implicit = (match tpcls with | IdConstant n -> n | _ -> 0) in
  try (
(* TODO: maybe we should match type instead? *)
  if n_implicit = 0
  then
    (
      match spine with
      | []  -> (ctx, spine, fed)
      | (m :: ms) -> (
              match fed with
              | Pi(x, t2, t1) -> (
                  let (ctx, cm, t2) = tr_check_type m t2 ctx lg in
                  let t1' = HS.h_sub_in_tp cm (A.to_weak t2) x t1 in
                  let (ctx, cms, final_t) = tr_feed_spine ms (t1', tpcls) ctx lg in
                  (ctx, (cm :: cms), final_t)
              )
              | Atp(A.Gen v, s) -> ( (* TODO consider free var of types *)
                let binder = A.Prvd (Uid.nextz()) in
                let (ctx, t2univ) = Ctx.new_metavar_tp ctx in
                let (ctx, t1univ) = Ctx.new_metavar_tp ctx in
                let t2 = A.Atp(A.Gen t2univ, s) in
                let t1 = A.Atp(A.Gen t1univ, s@[A.Atm(binder, [])]) in
                let s = Unif.check_spine_is_pattern ctx s lg in
                let resolution_body = A.Pi(binder, t2, t1) in
                let resolution = Ctx.abstract_over_tp ctx s resolution_body lg in
                let ctx = Ctx.resolve_metavar ctx v (Ctx.ResolvedTp resolution) in
                let (ctx, m, t2) = tr_check_type m t2 ctx lg in
                let m = HS.eta_expand m (A.to_weak t2) in
                let t1 = HS.h_sub_in_tp m (A.to_weak t2) binder t1 in
                let (ctx, ms, final_t) = tr_feed_spine ms (t1, tpcls) ctx lg in
                (ctx, (m :: ms), final_t)
              )
              | _ -> failwith "expecting pi type 63: The number of arguments (=spine) may be more than the number of Pi's in the type of head (=into) of the term "
      )
    )
  else tr_feed_spine ((A.Atm(A.Placeholder, []))::spine) (fed, IdConstant (n_implicit -1)) ctx lg
    ) with Failure s -> raise (Failure (s ^ "\n when feeding spine " ^ A.ds_spine spine ^ 
    " into  " ^ A.ds_tp fed ^ ""  ^ Ctx.ds_conditional_show_ctx ctx))

and tr_check_type (tm : A.tm) (tp : A.tp) (ctx : Ctx.ctxtype) (lg : A.lgsig) : Ctx.ctxtype * A.tm * A.tp = 
    let _ = if debug then print_endline ("tr_check_type on " ^ A.ds_tm tm ^ " of type " ^  A.ds_tp tp ^ " in ctx " ^ Ctx.ds_conditional_show_ctx ctx ) else () in
    try(
  match tp with
  | (A.Type _ | A.CoType _) -> failwith "tr76: cannot check type/cotype"
  | A.Pi(tx, tt2, tt1) -> 
      (match tm with
        | A.Lam(lx, body) -> 
          (let ((x, body, tt1)) = UnifyUtil.unify_bnd_tm_tp  (lx, body) (tx, tt1) in
        let (ctx, (tm, tp)) = Ctx.with_new_name ctx x tt2 (fun ctx -> 
            let (ctx, body, tt1) = tr_check_type body tt1 ctx lg in
            (ctx, ((A.Lam(x, body)), A.Pi(x, tt2, tt1))))
           in (ctx, tm, tp))
        | A.Atm(_) -> 
          tr_check_type (HS.eta_expand tm (A.to_weak tp)) tp ctx lg
      )
  | A.Atp(_) -> (
      match tm with
      | A.Atm(A.Placeholder, s) -> (
          let s' = Unif.check_spine_is_pattern ctx s lg in
          (* let (ctx, tm_tp) = Ctx.new_metavar_tp ctx  in *)
          (* only collect varts that are not provided by the spine*)
          let vars = List.filter (fun (v, _) -> not (List.mem v s')) (Ctx.collect_all_bound_vars ctx) in
          let ids = List.map(fun (v,_) -> v) vars @ s' in
          let tml = List.map (fun (v, t) -> HS.eta_expand (A.Atm(v, [])) (A.to_weak t) ) vars in
          (* let tp = A.Atp(A.Gen tm_tp, tml) in take the *)
          let tp' = Ctx.abstract_over_tp ctx ids tp lg in
          let (ctx, tm_uname) = Ctx.new_metavar_tm ctx tp' []  in (* TODO: use metavar_tm_with_abstractions*)
          let tm = (A.Atm(A.Gen(tm_uname), tml@s)) in
          tr_check_type tm tp ctx lg
          (* (ctx, A.Gen tm_uname, tml, (tp, IdUniVar))  *)
        )
      | _ ->
        (let (ctx, tm, syn_tp) = tr_synth_type tm ctx lg in
        let (ctx, syn_tp) = Unif.request_unify_tp ctx syn_tp tp lg in
        (ctx, tm, syn_tp))
  )
    ) with Failure s -> raise (Failure (s ^ "\n when checking that term " ^ A.ds_tm tm ^ " has type "  ^ A.ds_tp tp ^ "" ^ Ctx.ds_conditional_show_ctx ctx ))


and tr_synth_type (tm : A.tm)  (ctx : Ctx.ctxtype) (lg : A.lgsig) : Ctx.ctxtype * A.tm  * A.tp = 
    let _ = if debug then print_endline ("tr_synth_type on " ^ A.ds_tm tm ^ " in ctx " ^ Ctx.ds_conditional_show_ctx ctx ) else () in
    try(
        match tm with
        | A.Lam _ -> failwith "tr76: cannot syn on lam"
        | A.Atm(h, s) -> 
          let (ctx, h, resolvedtp) = Ctx.resolve_id ctx h lg in 
          let (ctx, s, k) =  tr_feed_spine s resolvedtp ctx lg in 
            (ctx, A.Atm(h, s), k)
    ) with Failure s -> raise (Failure (s ^ "\n when synthesizing the type of " ^ A.ds_tm tm ^ ""  ^ Ctx.ds_conditional_show_ctx ctx))


(* pending holds things that have been encountered but whose type is being abstracted*)
let rec abstract_implicit_tm (ctx : Ctx.ctxtype) (tm : A.tm) (tp : A.tp) (abstracted : (A.id * A.tp) list)(pending : A.id list) (lg : A.lgsig) 
    : (Ctx.ctxtype * (A.tm (* processed A.tp, may replace metavars with bound vars *)
                     * (A.id * A.tp) list (* list of bound variables (id remains gen if univar and prvd if freevar), [x1, x2] corresponds to \Pi x1. \Pi x2. *)
                      )) = 
    let _ = if debug then (print_endline ("abstract_implicit_tm abstracting term " ^ A.ds_tm tm ^ " of type " ^ A.ds_tp tp 
      ^ " abstracted is [" ^ String.concat ", " (List.map (fun (i,t) -> A.ds_id i ^ ": " ^ A.ds_tp t) abstracted)  ^ "]" ^
       " " ^ Ctx.ds_conditional_show_ctx ctx)) else () in
    try(
    let visited id = List.exists (fun (n, _) -> n = id) abstracted in
    match (tm,tp) with
    | (A.Lam(x, b), A.Pi(z,t2,t1)) -> 
        let (x, b, t1) = UnifyUtil.unify_bnd_tm_tp (x, b) (z, t1) in
        Ctx.with_new_name ctx x t2 (fun ctx -> 
            let (ctx, (b, abstracted)) = abstract_implicit_tm ctx b t1 abstracted pending lg in
            (ctx, (A.Lam(x, b), abstracted))
          ) 
    | (A.Atm(h, s), A.Atp _) -> (
          let h' = match h with | A.Prvd _ -> h | A.Gen s -> A.Prvd s | A.Placeholder -> failwith "tr171: no plhd" in
          let (tp, _) = Ctx.lookup_id_in_ctx_and_sig ctx true h' lg in
          let pending' = 
            if List.mem h pending then failwith "cycle detected during abstracting" else h :: pending in
          let (ctx, (tp, abstracted)) = abstract_implicit_tp ctx tp abstracted pending' lg in
          let (abstracted) =  (match Ctx.lookup_id_in_ctx_and_sig ctx false h lg with
            | (_, Ctx.IdBound) -> abstracted
            | (_, Ctx.IdConstant _) -> abstracted
            | (_, Ctx.IdRecConstantInCtx _) -> abstracted
            | (_, Ctx.IdRecConstantInSig _) -> abstracted
            | (_, Ctx.IdFreeVar) -> if visited h' then abstracted else ( abstracted @ [(h', tp)])
            | (_, Ctx.IdUniVar _) -> if visited h' then abstracted else ( abstracted @ [(h', tp)])
          ) in 
          let (ctx, (s, abstracted)) = abstract_implicit_spine ctx s tp abstracted pending lg in
          (ctx, (A.Atm(h', s), abstracted))
    )
    | _ -> failwith "tr173: type mismatch"
    ) with Failure s -> raise (Failure (s ^ "\n when abstracting term " ^ A.ds_tm tm ^ " of type " ^ A.ds_tp tp ^ " " ^ Ctx.ds_conditional_show_ctx ctx))
        
and abstract_implicit_spine (ctx : Ctx.ctxtype) (spine : A.tm list) (tp : A.tp) (abstracted : (A.id * A.tp) list)(pending : A.id list) (lg : A.lgsig) 
    : (Ctx.ctxtype * (A.tm list
                     * (A.id * A.tp) list 
                      )) = 
    let _ = if debug then print_endline ("abstract_implicit_spine abstracting spine " ^ A.ds_spine spine ^ " of type " ^ A.ds_tp tp 
      ^ " abstracted is [" ^ String.concat ", " (List.map (fun (i,t) -> A.ds_id i ^ ": " ^ A.ds_tp t) abstracted)  ^ "]" 
      ^ " " ^ Ctx.ds_conditional_show_ctx ctx) else () in
    try(
      match (spine, tp) with
      | ([], A.Atp(_)) -> (ctx, (spine, abstracted))
      | ([], A.Type _) -> (ctx, (spine, abstracted))
      | ([], A.CoType _) -> (ctx, (spine, abstracted))
      | ((m :: ms), A.Pi(x, t2, t1)) -> 
          ( let (ctx, (m, abstracted)) = abstract_implicit_tm ctx m t2 abstracted pending lg in
            let t2' = HS.h_sub_in_tp m (A.to_weak t2) x t1 in
            let (ctx, (ms, abstracted)) = abstract_implicit_spine ctx ms t2' abstracted pending lg in
            (ctx, ((m :: ms), abstracted))
          )
      | _ -> failwith ("tr163: abstract spine type mismatch when abstracting " ^ A.ds_spine spine ^ " of type " ^ A.ds_tp tp)
    ) with Failure s -> raise (Failure (s ^ "\n when abstracting spine " ^ A.ds_spine spine ^ " of type " ^ A.ds_tp tp ^ " " ^ Ctx.ds_conditional_show_ctx ctx))
              
and abstract_implicit_tp (ctx : Ctx.ctxtype) (tp : A.tp) (abstracted : (A.id * A.tp) list) (pending : A.id list) (lg : A.lgsig) 
    : (Ctx.ctxtype * (A.tp (* processed A.tp, may replace metavars with bound vars *)
                     * (A.id * A.tp) list (* list of bound variables (id remains gen if univar and prvd if freevar), [x1, x2] corresponds to \Pi x1. \Pi x2. *)
                      )) = 
    let _ = if debug then print_endline ("abstract_implicit_spine abstracting tp " ^ A.ds_tp tp 
      ^ " abstracted is [" ^ String.concat ", " (List.map (fun (i,t) -> A.ds_id i ^ ": " ^ A.ds_tp t) abstracted)  ^ "]" 
      ^ " " ^ Ctx.ds_conditional_show_ctx ctx) else () in
    try(
      let visited id = List.exists (fun (n, _) -> n = id) abstracted in
      match tp with
      | A.Pi(x, t2, t1) -> 
          let (ctx, (t2, abstracted)) = abstract_implicit_tp ctx t2 abstracted pending lg in
          Ctx.with_new_name ctx x t2 (fun ctx -> 
              let (ctx, (t1, abstracted)) = abstract_implicit_tp ctx t1 abstracted pending lg in
              (ctx, (A.Pi(x, t2, t1), abstracted))
            ) 
      | A.Atp(h, s) -> (
            let h' = match h with | A.Prvd _ -> h | A.Gen s -> A.Prvd s | A.Placeholder -> failwith "tr171: no plhd" in
            match Ctx.find_id_in_ctx_and_sig ctx true h' lg with
            | None -> failwith ("Unresolved Metavariable " ^ A.ds_id h ^ " Type-Level metavariables must be resolved")
            | Some (tp, _) ->
              (
                let pending' = 
                  if List.mem h pending then failwith "cycle detected during abstracting" else h :: pending in
                let (ctx, (tp, abstracted)) = abstract_implicit_tp ctx tp abstracted  pending' lg in
                let (abstracted) =  (match Ctx.lookup_id_in_ctx_and_sig ctx false h lg with
                  | (_, Ctx.IdBound) -> abstracted
                  | (_, Ctx.IdConstant _) -> abstracted
                  | (_, Ctx.IdRecConstantInSig _ ) -> abstracted
                  | (_, Ctx.IdRecConstantInCtx _ ) -> abstracted
                  | (_, Ctx.IdFreeVar) -> if visited h' then abstracted else ( abstracted @ [(h', tp)])
                  | (_, Ctx.IdUniVar _) -> if visited h' then abstracted else ( abstracted @ [(h', tp)])
                ) in 
                let (ctx, (s, abstracted)) = abstract_implicit_spine ctx s tp abstracted pending lg in
                (ctx, (A.Atp(h', s), abstracted))
              )
      )
      | A.Type p -> (ctx, (A.Type p, abstracted))
      | A.CoType p -> (ctx, (A.CoType p, abstracted))
    ) with Failure s -> raise (Failure (s ^ "\n when abstracting tp " ^ A.ds_tp tp ^ " " ^ Ctx.ds_conditional_show_ctx ctx))


(* append to the local signature the updated recursion constants, 
   along with a (fresh, old) renaming function for recursion constants *)
   (* does not change the passed in signature*)
let update_local_sig_with_def_from_ctx (ctx : Ctx.ctxtype) (l : A.dec list) : (A.id * A.id) list * A.dec list = 
  try(
  let lctx = Ctx.populate_ctx_with_rec_constants_from_signature [] l in
  let rec_consts = A.collect_all_rec_constants_with_checked_body l in
  let changed = Circular.definition_changed_for_rec_consts rec_consts lctx ctx in
  (* essentially we require them do not change, which sofar has been the case during type reconstruciton *)
    if changed = [] then ([], l) else failwith ("tr278: NOT IMPLEMENTED changed " ^ String.concat ", " (List.map (fun (f) -> A.ds_id f) changed)
    ^ " type reconstruction should not have altered recursive definitions, please report this signature to the author for investigation.  The author can be reached at zhiboc@andrew.cmu.edu.  ")
  ) with Failure s -> raise (Failure (s ^ "\n when updating local signatures with def from ctx " ^ Ctx.ds_conditional_show_ctx ctx))




(* also returns the updated local signature with possibly new rec constants *)
let check_kind_or_type (t : A.tp) ((l, _) as lg : A.lgsig) : A.dec list * A.tp * int = 
  let ctx = [] in
  let ctx = Ctx.populate_ctx_with_rec_constants_from_signature ctx l in
  let ctx = Ctx.populate_ctx_with_free_vars_from_tp ctx t lg in
  let (ctx, t) = (if A.is_kind t then tr_kind else tr_type_cotype) t ctx lg in
  let t = Ctx.instantiate_metavar_tp ctx t in (* TODO Can we remove this once metavars behave like references? *)
  let (ctx, (t, abstracted)) = abstract_implicit_tp ctx t [] [] lg in
  let res_tp = List.fold_right  (fun (z, tp) acc -> 
      A.Pi((match z with
      | A.Gen s -> A.Prvd s
      | _ -> z), tp, acc)
    ) abstracted t  in
  let (renamings, l) = update_local_sig_with_def_from_ctx ctx l in
  let res_tp = List.fold_right (fun (fresh, old) tp -> HS.rename_in_tp fresh old tp) renamings res_tp in
  (l, res_tp, List.length abstracted)


  (* also returns the updated local signature with possibly recursion constants *)
(* adds this checked declaration into the returning context*)
let check_def_body (name : A.id) (m : A.tm) (t : A.tp) (err_decl : A.dec) ((l,r) : A.lgsig) : (A.dec list ) = 
  try(
    let ctx = [] in
    let ctx = Ctx.populate_ctx_with_rec_constants_from_signature ctx l in
    let (ctx, m, _) = tr_check_type m t ctx (l,r) in
    let m = Ctx.instantiate_metavar_tm ctx m in (* TODO Can we remove this once metavars behave like references? *)
    (* let _ = Validity.check_guarded_in_tm [name] m (l@r) in *)
    let (renamings, l) = update_local_sig_with_def_from_ctx ctx l in
    let m = List.fold_right (fun (fresh, old) tm -> HS.rename_in_tm fresh old tm) renamings m in
    
    (* let _ = match  abstract_implicit_tm ctx m t  [] [] (l,r) with 
                | (_,(_, bound)) -> if List.length bound > 0 then failwith "definition contains unsolved holes" else () in *)

    let (ctx, m') = Circular.transform_into_contractive ctx m t (l, r) in
    let ctx  = Ctx.RecDef(name, t, m') :: ctx in
    let all_reachable = Circular.collect_all_reachable_rec_constants_in_ctx ctx [name] [] in
    let checked = A.Def(name, t, m, A.CheckedBody all_reachable ) in 
    (l@[checked])
  )
  with Failure s -> raise (Failure (s ^ "\n when checking the definition body of declaration " ^ A.ds_dec (err_decl)
  ^  "\nchecked signatures: " ^ A.ds_decl l)
  )


let  tr_decl (tr : A.dec) ((_, _) as lg : A.lgsig) : A.dec list * A.dec = 
  let _ = if debug then print_endline ("=======\ntr_decl on " ^ A.ds_dec tr ^ "\n=======") else () in
  try(
    match tr with
    | A.Decl(name, t, None) -> 
        let (l, t', n_imp) = check_kind_or_type t lg in 
        (l, A.Decl(name, t', Some n_imp))
    | A.Decl(_, _, Some _) -> failwith "tr_decl should have empty n_implicit (decl)"
    | A.Def(name, t, m, A.Unchecked) (* ignore tm in the first pass*) ->
        let (l, t', n_imp) = check_kind_or_type t lg in 
        if n_imp <> 0 then failwith "the type of definitions must be fully specified" else
        let (free_r_vars) = TypeCheckingContext.get_rec_constants_in_tm_sig m lg in
        (l, A.Def(name, t',m, A.CheckedTp free_r_vars))
    | A.Def(_,_, _, _) -> failwith "tr_decl should have Unchecked (def)"
  ) with Failure s -> raise (Failure (s ^ "\n when checking the declaration " ^ A.ds_dec tr  
  ^  "\nchecked signatures: " ^ A.ds_decl  (let (local, _) = lg in local)
  ))


(* must be called with local = [], and r = full to check full, returns checked full*)
let rec tr_check_rec_def  (checklist : A.id list) ((l,r) : A.lgsig)   : A.dec list = 
  (* let _ = if debug then print_endline ("=======\ntr_check_rec_def on " ^ A.ds_dec tr ^ "\n=======") else () in *)
  (
    match r with
    | (A.Decl(_, _, None) :: _) -> failwith "should have been assinged a number during the first pass"
    | (A.Decl(_, _, Some _) as h :: t) -> tr_check_rec_def checklist (l@[h], t)
    | (A.Def(_, _, _, A.CheckedBody _) as h :: t) -> tr_check_rec_def checklist (l@[h], t)
    | (A.Def(name, tp, tm, A.CheckedTp _) as h :: lt) -> 
      if List.mem name checklist 
        then (
            (* updates happen here *)
            let (l') = check_def_body name tm tp h (l, r) in 
            (* note that flattening occurs after the entire procedure of type checking, 
               an error in flattening will only be found when later some types uses the recursive definition, 
               not here
               *)
            (* let (ctx, m') = Circular.transform_into_contractive [] m tp (l, r) in
            let ctx  = Ctx.RecDef(name, tp, m') :: ctx in
            let all_reachable = collect_all_reachable_rec_constants_in_ctx ctx [name] [] in
            let checked = A.Def(name, tp, m, A.CheckedBody all_reachable ) in *)
            tr_check_rec_def checklist (l', lt)
          )
        else tr_check_rec_def checklist (l@[h], lt)
    | (A.Def(_, _,_, A.Unchecked) :: _) -> failwith "should have been checked type during the first pass"
    | [] -> l
  ) 

let rec collect_all_rec_constants (l : A.dec list) : A.id list = 
  match l with
  | [] -> []
  | (A.Decl(_, _, _) :: t) -> collect_all_rec_constants t
  | (A.Def(name, _, _, _) :: t) -> name :: collect_all_rec_constants t

(* collects all recursion constants that 
   have only checked types but no the body, that are hereditarily 
   reachable from encoutered, without repetition, 
   does not include recursion constants that have checked body
   raise an exception if encountered recursion constants that have not been type checked
   *)
let rec collect_all_type_checked_reachable_rec_constants_in_sig
    (g : A.dec list) (to_collect : A.id list) (collected : A.id list) : A.id list = 
    (* collected is a sublist of to_collect *)
  match to_collect with
  | [] -> collected
  | (A.Prvd x :: xs) -> (
      match Ctx.find_id_sig_global false x g  with
      | Some (_, Ctx.IdRecConstantInSig(_, A.CheckedTp newids)) -> 
        (* finds the rechable ids that have not yet been collected*)
        let newids = List.map (fun v -> A.Prvd(v)) newids in
        let not_collected = ListUtil.minus newids collected in
        let to_collect_diff = ListUtil.minus not_collected to_collect in
        let new_to_collect = to_collect_diff @ xs in
        collect_all_type_checked_reachable_rec_constants_in_sig g new_to_collect (A.Prvd x :: collected) 
      | Some (_, Ctx.IdRecConstantInSig(_, A.CheckedBody _)) -> 
        (* do not collect things that have been checked*)
        collect_all_type_checked_reachable_rec_constants_in_sig g (xs) collected
      | Some (_, Ctx.IdRecConstantInSig(_, A.Unchecked)) -> 
        failwith "encoutered unchecked recursion constants when traversing reachable recursion constants from the next type/def declaration"
      | _ -> failwith "trying to collect non-rec constants"
  )
  | _ -> failwith "Collecting Non Prvd names"

(* needs to be checked separately because of recursive definitions, 
   guardedness require all defs to be checked*)
let rec tr_check_rec_guarded  (checklist : A.id list) ((l,r) : A.lgsig)   : A.dec list = 
  (* let _ = if debug then print_endline ("=======\ntr_check_rec_def on " ^ A.ds_dec tr ^ "\n=======") else () in *)
  (
    match r with
    | (A.Decl(_, _, None) :: _) -> failwith "should have been assinged a number during the first pass"
    | (A.Decl(_, _, Some _) as h :: t) -> tr_check_rec_guarded checklist (l@[h], t)
    | (A.Def(name, _, tm, A.CheckedBody _) as h :: lt) -> 
      if List.mem name checklist 
        then (
            (* updates happen here *)
            let _ = Validity.check_guarded_in_tm name [] [] tm (l@r) in
            (* note that flattening occurs after the entire procedure of type checking, 
               an error in flattening will only be found when later some types uses the recursive definition, 
               not here
               *)
            (* let (ctx, m') = Circular.transform_into_contractive [] m tp (l, r) in
            let ctx  = Ctx.RecDef(name, tp, m') :: ctx in
            let all_reachable = collect_all_reachable_rec_constants_in_ctx ctx [name] [] in
            let checked = A.Def(name, tp, m, A.CheckedBody all_reachable ) in *)
            tr_check_rec_guarded checklist (l@[h], lt)
          )
        else tr_check_rec_guarded checklist (l@[h], lt)
    | (A.Def(name, _,_, st) as h :: lt) -> 
      if List.mem name checklist 
      then
            failwith ("should have been checked body and types during prior pass  but got  " ^ A.ds_rec_check_state st ^ " for " ^ A.ds_dec h
          ^ "\n when checking guradedness for " ^ A.ds_ids checklist
          ^ "\n checked signatures: " ^ A.ds_decl l 
          ^ "\n remaining : " ^ A.ds_decl r)
      else tr_check_rec_guarded checklist (l@[h], lt)
    | [] -> l
  ) 

let  tr_check_rec_def_and_guarded  (checklist : A.id list) (g : A.dec list)   : A.dec list = 
    let g = tr_check_rec_def checklist ([], g) in
    let g = tr_check_rec_guarded checklist ([], g) in
    g

    

(* this function inspects the type of next definition, and 
   checks the definition all recursion constants that are referenced
    in the type.
    If there is no next type to look at, this function checks all recursion constants 
    that have not been checked
  *)
let lazy_check_rec_def ((checked, remaining) as lg : A.lgsig) : A.lgsig = 
  match remaining with
  | [] -> (tr_check_rec_def_and_guarded (collect_all_rec_constants checked) (checked), [])
  | ((A.Def(_, tp, _, A.Unchecked) | A.Decl(_, tp, None) ) :: _)  -> 
      (
        let rec_vars_mentioned = Ctx.get_rec_constants_in_tp tp lg in
        (* TODO: maybe consider anonymous constants? *)
        let rec_vars_mentioned = List.map (fun v -> A.Prvd v)  rec_vars_mentioned in
        let all_to_be_checked = collect_all_type_checked_reachable_rec_constants_in_sig (checked@remaining) rec_vars_mentioned [] in
        let checked = tr_check_rec_def_and_guarded all_to_be_checked (checked) in
        (checked, remaining)
      )
  | _ -> failwith "lazy_check_rec_def called with checked remaining, should be called with unchecked remaining, check implementation"
  

(* checks signature, skips recursion definitions *)
let rec tr_sig ((localchecked , remaining) : A.lgsig) : A.dec list
= 
match remaining with
| [] -> localchecked
| (h::t) -> 
  let (localchecked, tlocal) = tr_decl h (localchecked, remaining) in 
  let lg = ((localchecked@[tlocal]), t) in 
  let lg = lazy_check_rec_def (lg) in
  tr_sig  lg

let assert_no_repeat_names (l : A.dec list) : A.id list = 
  List.fold_left (fun (occurred_names: A.id list) dec -> 
      match dec with
      |(A.Def(name, _, _, _) | A.Decl(name, _,_)) -> 
        if List.mem name occurred_names
          then failwith ("name " ^ A.ds_id name ^ " occurred twice")
      else occurred_names@[name]) [] l

let  typecheckSig (l : A.dec list) : A.dec list
= let _ = assert_no_repeat_names l in 
tr_sig ([], l)

(*
Comment: 

Type checking is stratified:
The body of recursion constants must be checked before 
they are used in types
*)