

type id = 
  Prvd of string (* provided *)
  | Gen of string (* generated metavariable name *)
  | Placeholder (* for _ * in source expression*)


type tm = 
  Lam of id * tm
  | Atm of id * tm list
and tp = 
  Pi of id  * tp * tp
  | Atp of id * tm list
  | Type  of int (* priority higher is higher *) (* contrary to CoLF1 *)
  | CoType of int (* priority higher is higher *) (* contrary to CoLF1 *)

type weaktp = 
  | Atom
  | WeakTpUnivar of string
  | Arr of weaktp * weaktp

type rec_sig = (id * tp * tm) list
type rec_check_state = 
  Unchecked
  | CheckedTp of string list (* a list of recursion definition referenced in the body*)
  | CheckedBody of rec_sig (* flattened list of definitions including itself *)
  (* the body of this definition is checked*)
type dec = 
  Decl of id * tp * (int option) (* number of implicit arguments*)
  | Def of id * tp * tm * (rec_check_state  (* stores a list of recursion constants that are 
     referenced in this. When populated, it means that the type of the definition has been checked, 
     and the recursion constants referenced in the body has been collected, 
     the body of the definition is not checked. *)
     
  (* used for marking whether the type has been checked , the int should always be zero *)
  ) 
     (* coinductive definitions only, inductive definitions come later *)
  (* | DefC of id * tp * tm *)

(* Reinterpret this as lr sig: global = l (local) + remaning*)
type lgsig = dec list * dec list


let placeholder_tp = (Atp(Prvd "__placeholder__", []))

let rec is_kind (t : tp) : bool = 
  match t with
  | Type _ -> true
  | CoType _ -> true
  | Atp _ -> false
  | Pi(_, _, t) -> is_kind t


let rec to_weak (t : tp) : weaktp = 
  match t with
  (* | Type | CoType -> failwith "cannot convert Type/CoType to weak" *)
  | Type _ | CoType _ -> Atom
  | Atp (Prvd _, _) -> Atom
  | Atp (Gen s, _) -> WeakTpUnivar s
  | Atp (Placeholder, _) -> failwith "ast46: cannot get weak type containing a placeholder"
  | Pi(_, t2, t1) -> Arr(to_weak t2, to_weak t1)

let rec ds_weaktp (t : weaktp): string = 
  match t with
  | Atom -> "*"
  | WeakTpUnivar s -> "[?" ^ s ^ "]"
  | Arr(t2,t1) -> ds_weaktp t2 ^ " -> " ^ ds_weaktp t1

let rec ds_id (a : id) : string = 
  match a with
  | Prvd s  -> s
  | Gen s -> s
  | Placeholder -> "_" 
and ds_ids (l : id list) : string = 
  String.concat ", " (List.map ds_id l)
and ds_tm (t : tm) : string = 
  match t with
  | Lam (x,m) -> " [ " ^ ds_id x ^  " ] " ^ ds_tm m
  | Atm (h, tml) -> ds_hs(h,tml)
and ds_hs ((h, s) : id * (tm list)) : string = 
match s with
| [] -> ds_id h
| _ -> ds_id h ^ " ⋅ "  ^ ds_spine s
and ds_spine (( s) :  (tm list)) : string = 
"(" ^  String.concat ", " (List.map ds_tm s) ^ ")"
and ds_tp (t : tp) : string = 
  match t with
  (* | Pi (Placeholder,t2,t1) -> "(" ^ ds_tp t2 ^ ") -> " ^  ds_tp t1 *)
  | Pi (x,t2 ,t1) -> " { " ^ ds_id x ^ " : " ^ ds_tp t2 ^ " } " ^ ds_tp t1
  | Atp (h, tml) -> ds_hs(h, tml)
  | Type i -> "type(" ^ Int.to_string i ^ ")"
  | CoType i -> "cotype(" ^ Int.to_string i ^ ")"

and ds_id_tp_tm ((id, tp, tm) : (id * tp * tm) ) = 
  ds_id id  ^ " : " ^ ds_tp tp ^ " = " ^ ds_tm tm
and ds_rec_check_state (s : rec_check_state) : string = 
match s with
  | Unchecked -> "(unchecked)"
  | CheckedTp _ -> "(tp_checked)"
  | CheckedBody _l -> "(body_checked as " ^ 
  (* String.concat "; " (List.map ds_id_tp_tm _l) ^ *)
  ")" 

and ds_dec (d : dec) : string = 
match d with
  | Decl(c, a, Some i) -> ds_id c ^ " (" ^ string_of_int i ^ ") : " ^ ds_tp a
  | Decl(c, a, None) -> ds_id c ^  " : " ^ ds_tp a
  | Def(c,t,m, s) -> ds_id c ^ " " ^ ds_rec_check_state s ^ " : " ^ ds_tp t ^ " = " ^ ds_tm m
  (* | DefC(c,t,m) -> ds_id c ^ " :: " ^ ds_tp t ^ " = " ^ ds_tm m *)
and ds_decl ( dl : dec list) : string = 
  String.concat ".\n" (List.map ds_dec dl)
    

  
  
let rec collect_all_rec_constants_with_checked_body (l : dec list) : id list = 
  match l with
  | [] -> []
  | (Def(name, _, _, CheckedBody _) :: t) -> 
      name :: collect_all_rec_constants_with_checked_body t
  | (_ :: t) -> collect_all_rec_constants_with_checked_body t
